From 34f591031dfb09ba7d49ded5a49d0b61413bdd8e Mon Sep 17 00:00:00 2001
From: Liu, Jinsong <jinsong.liu@intel.com>
Date: Tue, 18 May 2010 16:54:39 +0800
Subject: [PATCH] KVM vcpu hotplug seabios infrastructure

This patch is to support KVM vcpu hotplug feature.
It ported from KVM bochs bios, and update some points to KVM seabios:
1. build ssdt with 15/255 processor obj, which separately support Win2000 and other oses;
2. build madt accordingly, protect it as 'E820_NVS', and keep madt checksum and lapic flag correctly when cpu hotplug;
3. define Processor under \_SB instead of \_PR, so that it works better with latest linux like 2.6.32 later;
4. for Processor _MAT method, explicitly transfer it 'ToBuffer' so that os will not confused with 'Integer' which will result in parse failure;
5. for Processor _STA method, use 0x0 instead of 0x9 for cpu hot-remove, which matchs better with ACPI 4.0 spec, and works better with latest linux;
6. add Processor _EJ0 method, so that 'eject' appear under linux /sys, and triggered by uevent to make cpu hot-remove;

Signed-off-by: Liu, Jinsong <jinosng.liu@intel.com>
---
 Makefile          |   14 ++++++-
 src/acpi-dsdt.dsl |    4 +-
 src/acpi-ssdt.dsl |  116 +++++++++++++++++++++++++++++++++++++++++++++++++++++
 src/acpi.c        |   88 ++++++++++++++++------------------------
 src/acpi.h        |    5 ++
 src/mk_ssdt.c     |   24 +++++++++++
 6 files changed, 194 insertions(+), 57 deletions(-)
 create mode 100644 src/acpi-ssdt.dsl
 create mode 100644 src/mk_ssdt.c

diff --git a/Makefile b/Makefile
index b896b54..21cab59 100644
--- a/Makefile
+++ b/Makefile
@@ -194,7 +194,19 @@ src/acpi-dsdt.hex: src/acpi-dsdt.dsl
 	@echo "Compiling DSDT"
 	$(Q)cpp -P $< > $(OUT)acpi-dsdt.dsl.i
 	$(Q)iasl -tc -p $@ $(OUT)acpi-dsdt.dsl.i
-	$(Q)rm $(OUT)acpi-dsdt.dsl.i
+	$(Q)sed -i -e 's/AmlCode/acpi_dsdt/g' $@
+	$(Q)rm $(OUT)acpi-dsdt.dsl.i src/acpi-dsdt.aml
+
+####### ssdt build rules
+15.c anycpu.c: %.c: src/acpi-ssdt.dsl src/mk_ssdt.c
+	$(Q)$(CC) -o mk_$* src/mk_ssdt.c
+	head -n -2 $< > src/acpi-ssdt-$*.dsl.i
+	./mk_$* >> src/acpi-ssdt-$*.dsl.i
+	@echo "Compiling SSDT support cpus:" $*
+	$(Q)cpp -P src/acpi-ssdt-$*.dsl.i > $(OUT)acpi-ssdt-$*.dsl
+	$(Q)iasl -tc -p src/acpi-ssdt-$*.hex $(OUT)acpi-ssdt-$*.dsl
+	$(Q)sed -i -e 's/AmlCode/acpi_ssdt_$*/g' src/acpi-ssdt-$*.hex
+	$(Q)rm mk_$* src/acpi-ssdt-$*.dsl.i src/acpi-ssdt-$*.aml $(OUT)acpi-ssdt-$*.dsl
 
 ####### Generic rules
 clean:
diff --git a/src/acpi-dsdt.dsl b/src/acpi-dsdt.dsl
index cc31112..96834fa 100644
--- a/src/acpi-dsdt.dsl
+++ b/src/acpi-dsdt.dsl
@@ -700,9 +700,7 @@ DefinitionBlock (
             Return (0x01)
 
         }
-        Method(_L02) {
-            Return(0x01)
-        }
+        /* _L02 defined at ssdt for CPU notification */
         Method(_L03) {
             Return(0x01)
         }
diff --git a/src/acpi-ssdt.dsl b/src/acpi-ssdt.dsl
new file mode 100644
index 0000000..9084c7a
--- /dev/null
+++ b/src/acpi-ssdt.dsl
@@ -0,0 +1,116 @@
+/*
+ * Bochs/QEMU ACPI SSDT ASL definition
+ *
+ * Copyright (c) 2006 Fabrice Bellard
+ *
+ * This library is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU Lesser General Public
+ * License version 2 as published by the Free Software Foundation.
+ *
+ * This library is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * Lesser General Public License for more details.
+ *
+ * You should have received a copy of the GNU Lesser General Public
+ * License along with this library; if not, write to the Free Software
+ * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ */
+DefinitionBlock (
+    "acpi-ssdt.aml",    // Output Filename
+    "SSDT",             // Signature
+    0x01,               // SSDT Compliance Revision
+    "BXPC",             // OEMID
+    "BXSSDT",           // TABLE ID
+    0x1                 // OEM Revision
+    )
+{
+    Scope (\_GPE)
+    {
+        /*
+         * _L02 method for CPU notification
+         */
+        Method(_L02) {
+            Return(\_SB.PRSC())
+        }
+    }
+
+
+    Scope (\_SB)
+    {
+        /* MADT_INFO_PHYSICAL_ADDRESS == 0x514 */
+        OperationRegion(MINF, SystemMemory, 0x514, 12)
+        Field(MINF, DwordAcc, NoLock, Preserve)
+        {
+            MSUA, 32, /* MADT checksum address */
+            MAPA, 32, /* MADT LAPIC0 address */
+        }
+
+        OperationRegion(MSUM, SystemMemory, MSUA, 1)
+        Field(MSUM, ByteAcc, NoLock, Preserve)
+        {
+            MSU, 8    /* MADT checksum */
+        }
+
+        #define gen_processor(nr, name)                                       \
+        Processor (C##name, nr, 0x0000b010, 0x06) {                           \
+            Name (_HID, "ACPI0007")                                           \
+            OperationRegion(MATR, SystemMemory, Add(MAPA, Multiply(nr,8)), 8) \
+            Field (MATR, ByteAcc, NoLock, Preserve)                           \
+            {                                                                 \
+                MAT, 64                                                       \
+            }                                                                 \
+            Field (MATR, ByteAcc, NoLock, Preserve)                           \
+            {                                                                 \
+                Offset(4),                                                    \
+                FLG, 1                                                        \
+            }                                                                 \
+            Method(_MAT, 0) {                                                 \
+                Return(ToBuffer(MAT))                                         \
+            }                                                                 \
+            Method (_STA) {                                                   \
+                If (FLG) { Return(0xF) } Else { Return(0x0) }                 \
+            }                                                                 \
+            Method (_EJ0, 1, NotSerialized) {                                 \
+                Sleep (0xC8)                                                  \
+            }                                                                 \
+        }                                                                     \
+
+        #define gen_ntfy(nr)                        \
+        If (LEqual(Arg0, 0x##nr)) {                 \
+            If (LNotEqual(Arg1, \_SB.C##nr.FLG)) {  \
+                Store (Arg1, \_SB.C##nr.FLG)        \
+                If (LEqual(Arg1, 1)) {              \
+                    Notify(C##nr, 1)                \
+                    Subtract(\_SB.MSU, 1, \_SB.MSU) \
+                } Else {                            \
+                    Notify(C##nr, 3)                \
+                    Add(\_SB.MSU, 1, \_SB.MSU)      \
+                }                                   \
+            }                                       \
+        }
+
+        OperationRegion(PRST, SystemIO, 0xaf00, 32)
+        Field (PRST, ByteAcc, NoLock, Preserve)
+        {
+            PRS, 256
+        }
+
+        Method(PRSC, 0) {
+                Store(PRS, Local3)
+                Store(Zero, Local0)
+                While(LLess(Local0, 32)) {
+                        Store(Zero, Local1)
+                        Store(DerefOf(Index(Local3, Local0)), Local2)
+                        While(LLess(Local1, 8)) {
+                                NTFY(Add(Multiply(Local0, 8), Local1),
+                                                And(Local2, 1))
+                                ShiftRight(Local2, 1, Local2)
+                                Increment(Local1)
+                        }
+                        Increment(Local0)
+                }
+                Return(One)
+        }
+    }
+}
diff --git a/src/acpi.c b/src/acpi.c
index 0559443..1f7a519 100644
--- a/src/acpi.c
+++ b/src/acpi.c
@@ -12,6 +12,8 @@
 #include "pci_ids.h" // PCI_VENDOR_ID_INTEL
 #include "pci_regs.h" // PCI_INTERRUPT_LINE
 #include "paravirt.h"
+#include "memmap.h"
+#include "types.h"
 
 /****************************************************/
 /* ACPI tables init */
@@ -269,6 +271,8 @@ struct srat_memory_affinity
 } PACKED;
 
 #include "acpi-dsdt.hex"
+#include "acpi-ssdt-15.hex"
+#include "acpi-ssdt-anycpu.hex"
 
 static inline u16 cpu_to_le16(u16 x)
 {
@@ -300,7 +304,7 @@ build_fadt(int bdf)
 {
     struct fadt_descriptor_rev1 *fadt = malloc_high(sizeof(*fadt));
     struct facs_descriptor_rev1 *facs = memalign_high(64, sizeof(*facs));
-    void *dsdt = malloc_high(sizeof(AmlCode));
+    void *dsdt = malloc_high(sizeof(acpi_dsdt));
 
     if (!fadt || !facs || !dsdt) {
         warn_noalloc();
@@ -313,7 +317,7 @@ build_fadt(int bdf)
     facs->length = cpu_to_le32(sizeof(*facs));
 
     /* DSDT */
-    memcpy(dsdt, AmlCode, sizeof(AmlCode));
+    memcpy(dsdt, acpi_dsdt, sizeof(acpi_dsdt));
 
     /* FADT */
     memset(fadt, 0, sizeof(*fadt));
@@ -344,24 +348,42 @@ build_fadt(int bdf)
     return fadt;
 }
 
+/* BIOS use only area unit: transfer MADT info to DSDT */
+#define MADT_INFO_PHYSICAL_ADDRESS    0x514
+struct madt_info {
+    u32 madt_csum_addr;    /* address of MADT checksum */
+    u32 madt_lapic0_addr;  /* address of first MADT LAPIC struct */
+};
+
 static void*
 build_madt(void)
 {
+    struct madt_info *madt_info;
+    int lapic_num = MaxCountCPUs > 15 ? MAX_CPU_GENEOS : MAX_CPU_WIN2000;
+
     int madt_size = (sizeof(struct multiple_apic_table)
-                     + sizeof(struct madt_processor_apic) * MaxCountCPUs
+                     + sizeof(struct madt_processor_apic) * lapic_num
                      + sizeof(struct madt_io_apic)
                      + sizeof(struct madt_intsrcovr) * 16);
-    struct multiple_apic_table *madt = malloc_high(madt_size);
+    madt_size = ALIGN(madt_size, PAGE_SIZE);
+    struct multiple_apic_table *madt = memalign_high(PAGE_SIZE, madt_size);
     if (!madt) {
         warn_noalloc();
         return NULL;
     }
+    add_e820((u32)madt, madt_size, E820_NVS);
     memset(madt, 0, madt_size);
     madt->local_apic_address = cpu_to_le32(BUILD_APIC_ADDR);
     madt->flags = cpu_to_le32(1);
     struct madt_processor_apic *apic = (void*)&madt[1];
+
+    madt_info = (struct madt_info *)MADT_INFO_PHYSICAL_ADDRESS;
+    memset(madt_info, 0, sizeof(*madt_info));
+    madt_info->madt_csum_addr = (u32)&madt->checksum;
+    madt_info->madt_lapic0_addr = (u32)apic;
+
     int i;
-    for (i=0; i<MaxCountCPUs; i++) {
+    for (i=0; i<lapic_num; i++) {
         apic->type = APIC_PROCESSOR;
         apic->length = sizeof(*apic);
         apic->processor_id = i;
@@ -406,63 +428,23 @@ build_madt(void)
     return madt;
 }
 
-#define SSDT_SIGNATURE 0x54445353 // SSDT
 static void*
 build_ssdt(void)
 {
-    int acpi_cpus = MaxCountCPUs > 0xff ? 0xff : MaxCountCPUs;
-    // calculate the length of processor block and scope block
-    // excluding PkgLength
-    int cpu_length = 13 * acpi_cpus + 4;
+    void *ssdt = malloc_high(MaxCountCPUs > 15 ? 
+                             sizeof(acpi_ssdt_anycpu) : 
+                             sizeof(acpi_ssdt_15));
 
-    int length = sizeof(struct acpi_table_header) + 3 + cpu_length;
-    u8 *ssdt = malloc_high(length);
     if (! ssdt) {
         warn_noalloc();
         return NULL;
     }
 
-    u8 *ssdt_ptr = ssdt;
-    ssdt_ptr[9] = 0; // checksum;
-    ssdt_ptr += sizeof(struct acpi_table_header);
-
-    // build processor scope header
-    *(ssdt_ptr++) = 0x10; // ScopeOp
-    if (cpu_length <= 0x3e) {
-        /* Handle 1-4 CPUs with one byte encoding */
-        *(ssdt_ptr++) = cpu_length + 1;
-    } else {
-        /* Handle 5-314 CPUs with two byte encoding */
-        *(ssdt_ptr++) = 0x40 | ((cpu_length + 2) & 0xf);
-        *(ssdt_ptr++) = (cpu_length + 2) >> 4;
-    }
-    *(ssdt_ptr++) = '_'; // Name
-    *(ssdt_ptr++) = 'P';
-    *(ssdt_ptr++) = 'R';
-    *(ssdt_ptr++) = '_';
-
-    // build object for each processor
-    int i;
-    for (i=0; i<acpi_cpus; i++) {
-        *(ssdt_ptr++) = 0x5B; // ProcessorOp
-        *(ssdt_ptr++) = 0x83;
-        *(ssdt_ptr++) = 0x0B; // Length
-        *(ssdt_ptr++) = 'C';  // Name (CPUxx)
-        *(ssdt_ptr++) = 'P';
-        if ((i & 0xf0) != 0)
-            *(ssdt_ptr++) = (i >> 4) < 0xa ? (i >> 4) + '0' : (i >> 4) + 'A' - 0xa;
-        else
-            *(ssdt_ptr++) = 'U';
-        *(ssdt_ptr++) = (i & 0xf) < 0xa ? (i & 0xf) + '0' : (i & 0xf) + 'A' - 0xa;
-        *(ssdt_ptr++) = i;
-        *(ssdt_ptr++) = 0x10; // Processor block address
-        *(ssdt_ptr++) = 0xb0;
-        *(ssdt_ptr++) = 0;
-        *(ssdt_ptr++) = 0;
-        *(ssdt_ptr++) = 6;    // Processor block length
-    }
-
-    build_header((void*)ssdt, SSDT_SIGNATURE, ssdt_ptr - ssdt, 1);
+    /* SSDT */
+    if (MaxCountCPUs > 15)
+        memcpy(ssdt, acpi_ssdt_anycpu, sizeof(acpi_ssdt_anycpu));
+    else
+        memcpy(ssdt, acpi_ssdt_15, sizeof(acpi_ssdt_15));
 
     return ssdt;
 }
diff --git a/src/acpi.h b/src/acpi.h
index c3ae84c..86c7859 100644
--- a/src/acpi.h
+++ b/src/acpi.h
@@ -3,6 +3,11 @@
 
 #include "types.h" // u32
 
+/* Windows 2000 support max cpus to 15, otherwise will result in BSOD */
+/* Other os support more cpus, we define here support max cpus to 255 */
+#define MAX_CPU_WIN2000 15
+#define MAX_CPU_GENEOS  255
+
 void acpi_bios_init(void);
 u32 find_resume_vector(void);
 
diff --git a/src/mk_ssdt.c b/src/mk_ssdt.c
new file mode 100644
index 0000000..e8815fc
--- /dev/null
+++ b/src/mk_ssdt.c
@@ -0,0 +1,24 @@
+#include <stdio.h>
+#include "acpi.h"
+
+int main(int argc, char **argv)
+{
+    unsigned int i, max_cpus = MAX_CPU_GENEOS;
+
+    sscanf(argv[0], "%*[^0-9]%u", &max_cpus);
+
+    printf("\n");
+    for (i = 0; i < max_cpus; i++)
+        printf("        gen_processor(%d, %x)\n", i, i);
+
+    printf("\n");
+    printf("        Method (NTFY, 2) {\n");
+    for (i = 0; i < max_cpus; i++)
+        printf("        gen_ntfy(%x)\n", i);
+    printf("        Return(One)\n");
+    printf("        }\n");
+    printf("    }\n");
+    printf("}\n");
+
+    return 0;
+}
-- 
1.6.5.6

